@@ -1,3 +1,4 @@
+import itertools
 from abc import ABC, ABCMeta
 from collections.abc import Callable
 from functools import wraps
@@ -36,7 +37,10 @@ ACTUAL = HERE / "figures"
 TOL = 15
 DPI = 80
 
-RNG = np.random.default_rng(seed=42)
+
+def get_standard_RNG():
+    # we init from scratch each time to ensure same results in each test
+    return np.random.default_rng(seed=42)
 
 
 @pytest.fixture()
@@ -116,7 +120,10 @@ def test_sdata_multiple_images_with_table():
     instance_key = "instance_id"
     region_key = "annotated_region"
 
-    adata = AnnData(RNG.normal(size=(30, 10)), obs=pd.DataFrame(RNG.normal(size=(30, 3)), columns=["a", "b", "c"]))
+    adata = AnnData(
+        get_standard_RNG().normal(size=(30, 10)),
+        obs=pd.DataFrame(get_standard_RNG().normal(size=(30, 3)), columns=["a", "b", "c"]),
+    )
     adata.obs[instance_key] = list(range(3)) + list(range(7)) + list(range(20))
     adata.obs[region_key] = ["data1"] * 3 + ["data2"] * 7 + ["data3"] * 20
     table = TableModel.parse(
@@ -154,7 +161,7 @@ def test_sdata_multiple_images_diverging_dims():
 def sdata_blobs_shapes_annotated() -> SpatialData:
     """Get blobs sdata with continuous annotation of polygons."""
     blob = blobs()
-    blob["table"].obs["region"] = "blobs_polygons"
+    blob["table"].obs["region"] = pd.Categorical(["blobs_polygons"] * blob["table"].n_obs)
     blob["table"].uns["spatialdata_attrs"]["region"] = "blobs_polygons"
     blob.shapes["blobs_polygons"]["value"] = [1, 2, 3, 4, 5]
     return blob
@@ -237,23 +244,29 @@ def _get_images() -> dict[str, DataArray | DataTree]:
     out = {}
     dims_2d = ("c", "y", "x")
     dims_3d = ("z", "y", "x", "c")
-    out["image2d"] = Image2DModel.parse(RNG.normal(size=(3, 64, 64)), dims=dims_2d, c_coords=["r", "g", "b"])
+    out["image2d"] = Image2DModel.parse(
+        get_standard_RNG().normal(size=(3, 64, 64)), dims=dims_2d, c_coords=["r", "g", "b"]
+    )
     out["image2d_multiscale"] = Image2DModel.parse(
-        RNG.normal(size=(3, 64, 64)), scale_factors=[2, 2], dims=dims_2d, c_coords=["r", "g", "b"]
+        get_standard_RNG().normal(size=(3, 64, 64)), scale_factors=[2, 2], dims=dims_2d, c_coords=["r", "g", "b"]
+    )
+    out["image2d_xarray"] = Image2DModel.parse(
+        DataArray(get_standard_RNG().normal(size=(3, 64, 64)), dims=dims_2d), dims=None
     )
-    out["image2d_xarray"] = Image2DModel.parse(DataArray(RNG.normal(size=(3, 64, 64)), dims=dims_2d), dims=None)
     out["image2d_multiscale_xarray"] = Image2DModel.parse(
-        DataArray(RNG.normal(size=(3, 64, 64)), dims=dims_2d),
+        DataArray(get_standard_RNG().normal(size=(3, 64, 64)), dims=dims_2d),
         scale_factors=[2, 4],
         dims=None,
     )
-    out["image3d_numpy"] = Image3DModel.parse(RNG.normal(size=(2, 64, 64, 3)), dims=dims_3d)
+    out["image3d_numpy"] = Image3DModel.parse(get_standard_RNG().normal(size=(2, 64, 64, 3)), dims=dims_3d)
     out["image3d_multiscale_numpy"] = Image3DModel.parse(
-        RNG.normal(size=(2, 64, 64, 3)), scale_factors=[2], dims=dims_3d
+        get_standard_RNG().normal(size=(2, 64, 64, 3)), scale_factors=[2], dims=dims_3d
+    )
+    out["image3d_xarray"] = Image3DModel.parse(
+        DataArray(get_standard_RNG().normal(size=(2, 64, 64, 3)), dims=dims_3d), dims=None
     )
-    out["image3d_xarray"] = Image3DModel.parse(DataArray(RNG.normal(size=(2, 64, 64, 3)), dims=dims_3d), dims=None)
     out["image3d_multiscale_xarray"] = Image3DModel.parse(
-        DataArray(RNG.normal(size=(2, 64, 64, 3)), dims=dims_3d),
+        DataArray(get_standard_RNG().normal(size=(2, 64, 64, 3)), dims=dims_3d),
         scale_factors=[2],
         dims=None,
     )
@@ -265,27 +278,27 @@ def _get_labels() -> dict[str, DataArray | DataTree]:
     dims_2d = ("y", "x")
     dims_3d = ("z", "y", "x")
 
-    out["labels2d"] = Labels2DModel.parse(RNG.integers(0, 100, size=(64, 64)), dims=dims_2d)
+    out["labels2d"] = Labels2DModel.parse(get_standard_RNG().integers(0, 100, size=(64, 64)), dims=dims_2d)
     out["labels2d_multiscale"] = Labels2DModel.parse(
-        RNG.integers(0, 100, size=(64, 64)), scale_factors=[2, 4], dims=dims_2d
+        get_standard_RNG().integers(0, 100, size=(64, 64)), scale_factors=[2, 4], dims=dims_2d
     )
     out["labels2d_xarray"] = Labels2DModel.parse(
-        DataArray(RNG.integers(0, 100, size=(64, 64)), dims=dims_2d), dims=None
+        DataArray(get_standard_RNG().integers(0, 100, size=(64, 64)), dims=dims_2d), dims=None
     )
     out["labels2d_multiscale_xarray"] = Labels2DModel.parse(
-        DataArray(RNG.integers(0, 100, size=(64, 64)), dims=dims_2d),
+        DataArray(get_standard_RNG().integers(0, 100, size=(64, 64)), dims=dims_2d),
         scale_factors=[2, 4],
         dims=None,
     )
-    out["labels3d_numpy"] = Labels3DModel.parse(RNG.integers(0, 100, size=(10, 64, 64)), dims=dims_3d)
+    out["labels3d_numpy"] = Labels3DModel.parse(get_standard_RNG().integers(0, 100, size=(10, 64, 64)), dims=dims_3d)
     out["labels3d_multiscale_numpy"] = Labels3DModel.parse(
-        RNG.integers(0, 100, size=(10, 64, 64)), scale_factors=[2, 4], dims=dims_3d
+        get_standard_RNG().integers(0, 100, size=(10, 64, 64)), scale_factors=[2, 4], dims=dims_3d
     )
     out["labels3d_xarray"] = Labels3DModel.parse(
-        DataArray(RNG.integers(0, 100, size=(10, 64, 64)), dims=dims_3d), dims=None
+        DataArray(get_standard_RNG().integers(0, 100, size=(10, 64, 64)), dims=dims_3d), dims=None
     )
     out["labels3d_multiscale_xarray"] = Labels3DModel.parse(
-        DataArray(RNG.integers(0, 100, size=(10, 64, 64)), dims=dims_3d),
+        DataArray(get_standard_RNG().integers(0, 100, size=(10, 64, 64)), dims=dims_3d),
         scale_factors=[2, 4],
         dims=None,
     )
@@ -335,7 +348,7 @@ def _get_polygons() -> dict[str, GeoDataFrame]:
 
 def _get_shapes() -> dict[str, AnnData]:
     out = {}
-    arr = RNG.normal(size=(100, 2))
+    arr = get_standard_RNG().normal(size=(100, 2))
     out["shapes_0"] = ShapesModel.parse(arr, shape_type="Square", shape_size=3)
     out["shapes_1"] = ShapesModel.parse(arr, shape_type="Circle", shape_size=np.repeat(1, len(arr)))
 
@@ -349,10 +362,10 @@ def _get_points() -> dict[str, pa.Table]:
     out = {}
     for i, v in enumerate(var_names):
         name = f"{name}_{i}"
-        arr = RNG.normal(size=(100, 2))
+        arr = get_standard_RNG().normal(size=(100, 2))
         # randomly assign some values from v to the points
-        points_assignment0 = pd.Series(RNG.choice(v, size=arr.shape[0]))
-        points_assignment1 = pd.Series(RNG.choice(v, size=arr.shape[0]))
+        points_assignment0 = pd.Series(get_standard_RNG().choice(v, size=arr.shape[0]))
+        points_assignment1 = pd.Series(get_standard_RNG().choice(v, size=arr.shape[0]))
         annotations = pa.table(
             {"points_assignment0": points_assignment0, "points_assignment1": points_assignment1},
         )
@@ -367,13 +380,16 @@ def _get_table(
 ) -> AnnData:
     region_key = region_key or "annotated_region"
     instance_key = instance_key or "instance_id"
-    adata = AnnData(RNG.normal(size=(100, 10)), obs=pd.DataFrame(RNG.normal(size=(100, 3)), columns=["a", "b", "c"]))
+    adata = AnnData(
+        get_standard_RNG().normal(size=(100, 10)),
+        obs=pd.DataFrame(get_standard_RNG().normal(size=(100, 3)), columns=["a", "b", "c"]),
+    )
     adata.obs[instance_key] = np.arange(adata.n_obs)
     if isinstance(region, str):
         table = TableModel.parse(adata=adata, region=region, instance_key=instance_key)
     elif isinstance(region, list):
-        adata.obs[region_key] = RNG.choice(region, size=adata.n_obs)
-        adata.obs[instance_key] = RNG.integers(0, 10, size=(100,))
+        adata.obs[region_key] = get_standard_RNG().choice(region, size=adata.n_obs)
+        adata.obs[instance_key] = get_standard_RNG().integers(0, 10, size=(100,))
         table = TableModel.parse(adata=adata, region=region, region_key=region_key, instance_key=instance_key)
     else:
         table = TableModel.parse(adata=adata, region=region, region_key=region_key, instance_key=instance_key)
@@ -510,3 +526,29 @@ def get_sdata_with_multiple_images(request) -> sd.SpatialData:
         return sdata
 
     return _get_sdata_with_multiple_images
+
+
+@pytest.fixture
+def sdata_hexagonal_grid_spots():
+    """Create a hexagonal grid of points for testing visium_hex functionality."""
+    from shapely.geometry import Point
+    from spatialdata.models import ShapesModel
+
+    spacing = 10.0
+    n_rows, n_cols = 4, 4
+
+    points = []
+    for i, j in itertools.product(range(n_rows), range(n_cols)):
+        # Offset every second row by half the spacing for proper hexagonal packing
+        x = j * spacing + (i % 2) * spacing / 2
+        y = i * spacing * 0.866  # sqrt(3)/2 for proper hexagonal spacing
+        points.append(Point(x, y))
+
+    # Create GeoDataFrame with radius column
+    gdf = GeoDataFrame(geometry=points)
+    gdf["radius"] = 2.0  # Small radius for original circles
+
+    # Use ShapesModel.parse() to create a properly validated GeoDataFrame
+    shapes_gdf = ShapesModel.parse(gdf)
+
+    return SpatialData(shapes={"spots": shapes_gdf})
