@@ -1,7 +1,6 @@
 import dask.array as da
 import matplotlib
 import matplotlib.pyplot as plt
-import numpy as np
 import pandas as pd
 import pytest
 import scanpy as sc
@@ -12,9 +11,8 @@ from spatialdata import SpatialData, deepcopy, get_element_instances
 from spatialdata.models import Labels2DModel, TableModel
 
 import spatialdata_plot  # noqa: F401
-from tests.conftest import DPI, PlotTester, PlotTesterMeta, _viridis_with_under_over
+from tests.conftest import DPI, PlotTester, PlotTesterMeta, _viridis_with_under_over, get_standard_RNG
 
-RNG = np.random.default_rng(seed=42)
 sc.pl.set_rcParams_defaults()
 sc.set_figure_params(dpi=DPI, color_map="viridis")
 matplotlib.use("agg")  # same as GitHub action runner
@@ -33,12 +31,12 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         sdata_blobs.pl.render_labels(element="blobs_labels").pl.show()
 
     def test_plot_can_render_multiscale_labels(self, sdata_blobs: SpatialData):
-        sdata_blobs["table"].obs["region"] = "blobs_multiscale_labels"
+        sdata_blobs["table"].obs["region"] = pd.Categorical(["blobs_multiscale_labels"] * sdata_blobs["table"].n_obs)
         sdata_blobs["table"].uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
         sdata_blobs.pl.render_labels("blobs_multiscale_labels").pl.show()
 
     def test_plot_can_render_given_scale_of_multiscale_labels(self, sdata_blobs: SpatialData):
-        sdata_blobs["table"].obs["region"] = "blobs_multiscale_labels"
+        sdata_blobs["table"].obs["region"] = pd.Categorical(["blobs_multiscale_labels"] * sdata_blobs["table"].n_obs)
         sdata_blobs["table"].uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
         sdata_blobs.pl.render_labels("blobs_multiscale_labels", scale="scale1").pl.show()
 
@@ -50,7 +48,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         img.attrs["transform"] = sdata_blobs["blobs_labels"].transform
         sdata_blobs["blobs_giant_labels"] = img
 
-        sdata_blobs["table"].obs["region"] = "blobs_giant_labels"
+        sdata_blobs["table"].obs["region"] = pd.Categorical(["blobs_giant_labels"] * sdata_blobs["table"].n_obs)
         sdata_blobs["table"].uns["spatialdata_attrs"]["region"] = "blobs_giant_labels"
 
         sdata_blobs.pl.render_labels("blobs_giant_labels").pl.show()
@@ -63,7 +61,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         img.attrs["transform"] = sdata_blobs["blobs_labels"].transform
         sdata_blobs["blobs_giant_labels"] = img
 
-        sdata_blobs["table"].obs["region"] = "blobs_giant_labels"
+        sdata_blobs["table"].obs["region"] = pd.Categorical(["blobs_giant_labels"] * sdata_blobs["table"].n_obs)
         sdata_blobs["table"].uns["spatialdata_attrs"]["region"] = "blobs_giant_labels"
 
         sdata_blobs.pl.render_labels("blobs_giant_labels", scale="full").pl.show()
@@ -110,7 +108,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
             max_col = max_col.str.replace("channel_", "ch").str.replace("_sum", "")
             max_col = pd.Categorical(max_col, categories=set(max_col), ordered=True)
             adata.obs["which_max"] = max_col
-            adata.obs["region"] = label
+            adata.obs["region"] = pd.Categorical([label] * adata.n_obs)
             del adata.uns["spatialdata_attrs"]
             table = TableModel.parse(
                 adata=adata,
@@ -142,7 +140,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         sdata_blobs_local = deepcopy(sdata_blobs)
 
         table = sdata_blobs_local["table"].copy()
-        table.obs["region"] = "blobs_multiscale_labels"
+        table.obs["region"] = pd.Categorical(["blobs_multiscale_labels"] * table.n_obs)
         table.uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
         table = table[:, ~table.var_names.isin(["channel_0_sum"])]
         sdata_blobs_local["multi_table"] = table
@@ -187,7 +185,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
 
     def test_can_plot_with_one_element_color_table(self, sdata_blobs: SpatialData):
         table = sdata_blobs["table"].copy()
-        table.obs["region"] = "blobs_multiscale_labels"
+        table.obs["region"] = pd.Categorical(["blobs_multiscale_labels"] * table.n_obs)
         table.uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
         table = table[:, ~table.var_names.isin(["channel_0_sum"])]
         sdata_blobs["multi_table"] = table
@@ -196,9 +194,9 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         ).pl.show()
 
     def test_plot_subset_categorical_label_maintains_order(self, sdata_blobs: SpatialData):
-        max_col = sdata_blobs.table.to_df().idxmax(axis=1)
-        max_col = pd.Categorical(max_col, categories=sdata_blobs.table.to_df().columns, ordered=True)
-        sdata_blobs.table.obs["which_max"] = max_col
+        max_col = sdata_blobs["table"].to_df().idxmax(axis=1)
+        max_col = pd.Categorical(max_col, categories=sdata_blobs["table"].to_df().columns, ordered=True)
+        sdata_blobs["table"].obs["which_max"] = max_col
 
         _, axs = plt.subplots(nrows=1, ncols=2, layout="tight")
 
@@ -210,9 +208,9 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         ).pl.show(ax=axs[1])
 
     def test_plot_subset_categorical_label_maintains_order_when_palette_overwrite(self, sdata_blobs: SpatialData):
-        max_col = sdata_blobs.table.to_df().idxmax(axis=1)
-        max_col = pd.Categorical(max_col, categories=sdata_blobs.table.to_df().columns, ordered=True)
-        sdata_blobs.table.obs["which_max"] = max_col
+        max_col = sdata_blobs["table"].to_df().idxmax(axis=1)
+        max_col = pd.Categorical(max_col, categories=sdata_blobs["table"].to_df().columns, ordered=True)
+        sdata_blobs["table"].obs["which_max"] = max_col
 
         _, axs = plt.subplots(nrows=1, ncols=2, layout="tight")
 
@@ -229,11 +227,11 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         instances = get_element_instances(sdata_blobs[labels_name])
         n_obs = len(instances)
         adata = AnnData(
-            RNG.normal(size=(n_obs, 10)),
-            obs=pd.DataFrame(RNG.normal(size=(n_obs, 3)), columns=["a", "b", "c"]),
+            get_standard_RNG().normal(size=(n_obs, 10)),
+            obs=pd.DataFrame(get_standard_RNG().normal(size=(n_obs, 3)), columns=["a", "b", "c"]),
         )
         adata.obs["instance_id"] = instances.values
-        adata.obs["category"] = RNG.choice(["a", "b", "c"], size=adata.n_obs)
+        adata.obs["category"] = get_standard_RNG().choice(["a", "b", "c"], size=adata.n_obs)
         adata.obs["category"][:3] = ["a", "b", "c"]
         adata.obs["region"] = labels_name
         table = TableModel.parse(
@@ -262,7 +260,7 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         ).pl.show()
 
     def test_plot_can_annotate_labels_with_table_layer(self, sdata_blobs: SpatialData):
-        sdata_blobs["table"].layers["normalized"] = RNG.random(sdata_blobs["table"].X.shape)
+        sdata_blobs["table"].layers["normalized"] = get_standard_RNG().random(sdata_blobs["table"].X.shape)
         sdata_blobs.pl.render_labels("blobs_labels", color="channel_0_sum", table_layer="normalized").pl.show()
 
     def _prepare_small_labels(self, sdata_blobs: SpatialData) -> SpatialData:
@@ -299,3 +297,25 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
         sdata_blobs = self._prepare_small_labels(sdata_blobs)
 
         sdata_blobs.pl.render_labels("blobs_labels_large", color="category", table_name="table").pl.show()
+
+
+def test_warns_when_table_does_not_annotate_element(sdata_blobs: SpatialData):
+    # Work on an independent copy since we mutate tables
+    sdata_blobs_local = deepcopy(sdata_blobs)
+
+    # Create a table that annotates a DIFFERENT element than the one we will render
+    other_table = sdata_blobs_local["table"].copy()
+    other_table.obs["region"] = "blobs_multiscale_labels"
+    other_table.uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
+    sdata_blobs_local["other_table"] = other_table
+
+    # Rendering "blobs_labels" with a table that annotates "blobs_multiscale_labels"
+    # should raise a warning and fall back to using no table.
+    with pytest.warns(UserWarning, match="does not annotate element"):
+        (
+            sdata_blobs_local.pl.render_labels(
+                "blobs_labels",
+                color="channel_0_sum",
+                table_name="other_table",
+            ).pl.show()
+        )
